from .utils import get_topologies

# ML methods
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import RandomForestRegressor, GradientBoostingRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.svm import SVR
import catboost as cb, xgboost as xg

# Autoregressive methods
from statsmodels.tsa.arima.model import ARIMA

# NN methods
from tensorflow.keras import Sequential, models
from tensorflow.keras.layers import LSTM, Dense, GRU, SimpleRNN, Conv1D, Flatten

# Metrics
from sklearn.metrics import mean_squared_error, mean_absolute_error, mean_absolute_percentage_error, r2_score
import numpy as np

class Model:
    '''
    Represents generic model with unified interface.

    Backends: 

        ML: CatBoost, XGBoost, Sklearn;

        AR: Statsmodels;

        NN:TensorFlow.Keras.

    Models:

        ML: LinearRegression, DecisionTreeRegressor, RandomForestRegressor, GradientBoostingRegressor, SVR, CatBoostRegressor, XGBoostRegressor, XGBoostRandomForestRegressor;

        AR: ARMA, ARIMA;

        NN: CNN + SimpleRNN/LSTM/GRU, LSTM x1-3, GRU x1-3, SimpleRNN x1-3, CNN, MLP(1-3).


    Methods:

        fit(x_train, y_train) - fits model using x_train and y_train data;

        predict(x_train, y_train, x_test, x_test) - forecasts values using x_test;

        evaluate(x_test, x_test) - forecasts values using x_test, then measures MSE, MAE, MAPE and R2;

        summary() - shows short summary about model, its type and backend.

    '''

    __type = ''
    __backend = ''
    __model_name = ''
    __model = None

    def __init__(self, model_name, lag = 15, data = None, optimizer = 'nadam', loss = 'mse'):
        models = {'ml': ['LR', 'DTR', 'RFR', 'GBR', 'SVR', 'CBR', 'XGBR', 'XGBRFR'],
                  'ar': ['ARMA(2,1)', 'ARIMA(2,1,1)'], 
                  'nn': ['CNN + LSTM', 'LSTM x3', 'LSTM x2', 'LSTM x1', 'CNN + GRU', 'GRU x3', 'GRU x2', 'GRU x1', \
                         'CNN + SimpleRNN', 'SimpleRNN x3', 'SimpleRNN x2', 'SimpleRNN x1', 'CNN', 'MLP(3)', 'MLP(2)', 'MLP(1)']}
        
        self.__type = [key for key,value in models.items() if model_name in value][0]
        self.__model_name = model_name

        if self.__type == 'ml':
            if model_name == 'CBR':
                self.__backend = 'Catboost'
            elif model_name[:2] == 'XG':
                self.__backend = 'XGBoost'
            else:
                self.__backend = 'Scikit-learn'
    
            self.__model = {  'LR' :  LinearRegression(),
                            'DTR' : DecisionTreeRegressor(min_samples_leaf=5),
                            'RFR' : RandomForestRegressor(),
                            'GBR' : GradientBoostingRegressor(),
                            'SVR' : SVR(kernel='linear', epsilon=1e-3),
                            'CBR' : cb.CatBoostRegressor(loss_function='MAPE'),
                            'XGBR': xg.XGBRegressor(objective='reg:squarederror'),
                            'XGBRFR':xg.XGBRFRegressor(objective = 'reg:squarederror')}[model_name]
        
        elif self.__type == 'ar':
            self.__backend = 'Statsmodels'
            if self.__model_name.find('I') != -1:
                p = float(self.__model_name[self.__model_name.find('(')+1:self.__model_name.find(',')])
                d = float(self.__model_name[self.__model_name.find(',')+1:self.__model_name.find(',', self.__model_name.find(',')+1)])
                q = float(self.__model_name[self.__model_name.find(',', self.__model_name.find(',')+1)+1:self.__model_name.find(')')])
            else:
                p = float(self.__model_name[self.__model_name.find('(')+1:self.__model_name.find(',')])
                d = 0
                q = float(self.__model_name[self.__model_name.find(',')+1:self.__model_name.find(')')])
            self.__model = ARIMA(data, order = (p, d, q), enforce_stationarity=False)
        
        elif self.__type == 'nn':
            self.__backend = 'TensorFlow.Keras'
            self.__model = Sequential(get_topologies(lag)[model_name])
            self.__model.compile(optimizer = optimizer, loss = loss,  metrics = ['mse', 'mae', 'mape'])

    def fit(self, x_train, y_train, epochs = 25, batch_size = 32):
        '''
        Fits model using x_train and y_train.
        
        Parameters:

        x_train - previous price values;

        y_train - target price;

        epochs - number of epochs in case of using neural network;

        batch_size - number of objects in one fit batch in case of using neural network.
        '''

        if self.__type == 'nn':
            x_train = x_train.reshape((x_train.shape[0], x_train.shape[1], 1))
            self.__model.fit(x_train, y_train, epochs = epochs, batch_size = batch_size, verbose = 0)
        elif self.__model_name == 'CBR':
            x_train = x_train.reshape((x_train.shape[0], x_train.shape[1]))
            self.__model.fit(x_train, y_train, silent = True)
        elif self.__backend != 'statsmodels':
            x_train = x_train.reshape((x_train.shape[0], x_train.shape[1]))
            self.__model.fit(x_train, y_train)
        
        return self

    def predict(self, x_test, scaler):
        '''
        Predicts and scales values using x_test and scaler.
        
        Args:

        x_test - previous price values;

        scaler - scaler will be used to scale values back.

        Result:

        predicted and scaled price values.
        '''

        if self.__type == 'ar':
            x_test = x_test.reshape((x_test.shape[0], x_test.shape[1]))
            return scaler.unscale(self.__model.fit().forecast(steps = 1)[0])
        elif self.__type == 'nn':
            x_test = x_test.reshape((x_test.shape[0], x_test.shape[1], 1))
            return scaler.unscale(self.__model.predict(x_test, verbose = 0).reshape(-1))
        else:
            x_test = x_test.reshape((x_test.shape[0], x_test.shape[1]))
            return scaler.unscale(self.__model.predict(x_test))

    def evaluate(self, x_test, y_test, scaler):
        '''
        Predicts and scales values using x_test and scaler, then measures MSE, MAE, MAPE and R2.
        
        Parameters:

        x_test - previous price values;

        scaler - scaler will be used to scale values back.

        Result:

        dict with errors values.
        '''
                
        prediction = self.predict(x_test, scaler)
        true = scaler.unscale(y_test)
        return {'mse': mean_squared_error(np.array(true), np.array(prediction)),
                'mae': mean_absolute_error(np.array(true), np.array(prediction)),
                'mape': mean_absolute_percentage_error(np.array(true), np.array(prediction)),
                'r2':  r2_score(np.array(true), np.array(prediction))}

    def summary(self):
        '''
        Shows short summary about model, its type and backend.
        '''
        
        print(f'{self.__type.upper()} model {self.__model_name}, backend is based on {self.__backend}.')


import pandas as pd, re, json, requests as rq, numpy as np, datetime

from .scaler import Scaler

from pathlib import Path
PACKAGEDIR = Path(__file__).parent.absolute()


def get_data(ticker = 'AAPL', column = 'close', length = 15, step = 1, start_date = '01-01-2013', split_date = '01-01-2022', end_date = '01-01-2023', flatten = False, validate = False):
    '''
    Prepares data for training and evaluating.

    Args:
    ticker (str): Ticker to get data for.
    column (str): Column from HLOCV to work with.
    length (int): Length of arrays, length-1 values are for fitting, 1 value is for evaluating.
    step (int): Step to cut data using sliding window.
    start_date (str): Date to start cutting data from.
    split_date (str): Date to split into train/test.
    end_date (str): Date to end cutting data.
    flatten (bool): Shape format:  (cuts, length-1) or (cuts, length-1, 1).
    validate (bool): To make validation split or not.
    
    Returns:
        x_train, x_test, y_train, y_test, scaler: train and test data with scaler to work with.

    Example:
    >>> get_data('AAPL', 'close', 10)
    [[0.122, 0.12, 0.125, ..], ..]
    [0.12, ..]
    [[0.132, 0.13, 0.135, ..], ..]
    [0.13, ..]
    Scaler()
    '''
    data = pd.read_parquet(f'{PACKAGEDIR}/data.parquet.gz')
    data = data[(data.dateTime >= start_date) & (data.ticker == ticker)]

    train_data = data[(data.dateTime < split_date)][column].values.reshape(-1, 1)
    test_data =  data[(data.dateTime >= split_date) & (data.dateTime < end_date)][column].values.reshape(-1, 1)

    train_set = [train_data[j:j+length,0] for j in range(0, len(train_data)-length, step)]
    test_set =  [test_data[j:j+length,0] for j in range(0, len(test_data)-length, step)]

    scaler = Scaler().fit(train_set)
    scaled_train_set = scaler.scale(train_set)
    scaled_test_set = scaler.scale(test_set)

    scaled_x_train, scaled_x_test = scaled_train_set[:,:length-1], scaled_test_set[:,:length-1]
    scaled_y_train, scaled_y_test = scaled_train_set[:,length-1],  scaled_test_set[:,length-1]

    if not flatten:
        scaled_x_train = np.reshape(scaled_x_train, (scaled_x_train.shape[0], length - 1,1))
        scaled_x_test = np.reshape(scaled_x_test, (scaled_x_test.shape[0], length - 1,1))

    if validate:
        val_data = data[(data.date >= end_date) & (data.ticker == ticker)]['price'].values.reshape(-1, 1)
        val_set = [val_data[j:j+length,0] for j in range(0, len(val_data)-length, step)]
        scaled_val_set = scaler.scale(val_set)
        scaled_x_val, scaled_y_val = scaled_val_set[:,:length-1], scaled_val_set[:,length-1]
        return scaled_x_train, scaled_x_test, scaled_x_val, scaled_y_train, scaled_y_test, scaled_y_val, scaler
    else:
        return scaled_x_train, scaled_x_test, scaled_y_train, scaled_y_test, scaler

def update_data(tickers = []):
    '''
    Updates stored data for defined tickers.

    Args:
    tickers (list): Tickers list to update data for.

    Example:
    >>> get_data(['AMZN', 'TSLA', 'NVDA']])
    Info about 3 tickers successfully uploaded.
    '''
    all_tickers_df = pd.read_parquet(f'{PACKAGEDIR}/data.parquet.gz')
    for i in tickers:
        headers = {"Accept":"text/html", "Accept-Language":"en-US", "Referer":"https://www.nasdaq.com/", "User-Agent":"Chrome/64.0.3282.119"} 
        resp = rq.get(f'https://api.nasdaq.com/api/quote/{i}/chart?assetclass=stocks&fromdate=2023-04-29&todate={datetime.datetime.now().strftime("%Y-%m-%d")}', headers=headers, verify=True)
        if resp.status_code == 200:
            try:
                smth = json.loads(re.search('\[.*\]', resp.text).group())
                cur_tick_data = pd.DataFrame([smth[k]['z'] for k in range(len(smth))])
                cur_tick_data['ticker'] = i
                for col_name in ['high','low','open','close','volume','value']:
                    cur_tick_data[col_name] = pd.to_numeric(cur_tick_data[col_name].str.replace(',',''))
                cur_tick_data['dateTime'] = pd.to_datetime(cur_tick_data['dateTime'])
                all_tickers_df = pd.concat([all_tickers_df, cur_tick_data])
            except KeyError:
                pass
        else:
            print(f'ERROR: smth is wrong with {i}')
    all_tickers_df.to_parquet(f'{PACKAGEDIR}/data.parquet.gz', compression='gzip')
    print(f'Info about {len(tickers)} tickers successfully uploaded.')


from tensorflow.keras import Sequential, models
from tensorflow.keras.layers import LSTM, Dense, GRU, SimpleRNN, Conv1D, Flatten

def get_topologies(lag = 30):
    '''
    Returns dict of neural network topologies adapted to defined lag. 

    Args:
        lag (int): 

    
    '''
    neurons_per_layer = lag-1
    input_shape = (neurons_per_layer, 1)
    return {  'CNN + LSTM': [Conv1D(filters = 128, kernel_size = 3, activation = 'relu', input_shape = input_shape),
                                LSTM(units = neurons_per_layer),
                                Dense(units = 1)],
                    'LSTM x3': [LSTM(units = neurons_per_layer, 
                                    return_sequences = True, 
                                    input_shape = input_shape),
                                LSTM(units = neurons_per_layer, 
                                    return_sequences = True),
                                LSTM(units = neurons_per_layer),
                                Dense(units = 1)],
                    'LSTM x2': [LSTM(units = neurons_per_layer, 
                                    return_sequences = True, 
                                    input_shape = input_shape),
                                LSTM(units = neurons_per_layer),
                                Dense(units = 1)],
                    'LSTM x1': [LSTM(units = neurons_per_layer, 
                                    input_shape = input_shape),
                                Dense(units = 1)],
                    'CNN + GRU': [Conv1D(filters = 128, kernel_size = 3, activation = 'relu', input_shape = input_shape),
                                GRU(units = neurons_per_layer),
                                Dense(units = 1)],
                    'GRU x3' : [GRU(units = neurons_per_layer, 
                                    return_sequences = True, 
                                    input_shape = input_shape),
                                GRU(units = neurons_per_layer, 
                                    return_sequences = True),
                                GRU(units = neurons_per_layer),
                                Dense(units = 1)],
                    'GRU x2' : [GRU(units = neurons_per_layer, 
                                    return_sequences = True, 
                                    input_shape = input_shape),
                                GRU(units = neurons_per_layer),
                                Dense(units = 1)],
                    'GRU x1' : [GRU(units = neurons_per_layer, 
                                    input_shape = input_shape),
                                Dense(units = 1)],
                    'CNN + SimpleRNN': [Conv1D(filters = 128, kernel_size = 3, activation = 'relu', input_shape = input_shape),
                                        SimpleRNN(units = neurons_per_layer),
                                        Dense(units = 1)],
                    'SimpleRNN x3':[SimpleRNN(units = neurons_per_layer, 
                                            return_sequences = True, 
                                            input_shape = input_shape),
                                    SimpleRNN(units = neurons_per_layer, 
                                            return_sequences = True),
                                    SimpleRNN(units = neurons_per_layer),
                                    Dense(units = 1)],
                    'SimpleRNN x2':[SimpleRNN(units = neurons_per_layer, 
                                            return_sequences = True, 
                                            input_shape = input_shape),
                                    SimpleRNN(units = neurons_per_layer),
                                    Dense(units = 1)],
                    'SimpleRNN x1':[SimpleRNN(units = neurons_per_layer, 
                                            input_shape = input_shape),
                                    Dense(units = 1)],
                    'CNN': [Conv1D(filters = 32, kernel_size = 5, input_shape = input_shape, activation = 'relu'),
                            Flatten(),
                            Dense(units = 1)],
                    'MLP(3)': [Dense(units = neurons_per_layer, input_shape = (neurons_per_layer,)),
                            Dense(units = neurons_per_layer*2),
                            Dense(units = neurons_per_layer),
                            Dense(units = 1)],        
                    'MLP(2)': [Dense(units = neurons_per_layer, input_shape = (neurons_per_layer,)),
                            Dense(units = neurons_per_layer),
                            Dense(units = 1)],
                    'MLP(1)': [Dense(units = neurons_per_layer, input_shape = (neurons_per_layer,)),
                            Dense(units = 1)]
                    }


import sys
sys.path.insert(0, '../ftsf')

import ftsf, pytest

def test_scaling_on_1d_array():
    assert ([[0,0.5,1,1.5]] == ftsf.CustomScaler().fit([[0,1,2,3]]).scale([[0,1,2,3]])).all()
def test_check_work1():
    assert 1 == 1
def test_check_work2():
    assert 1 == 1
def test_check_work3():
    assert 1 == 1


# eval 
# ключи и их порядок
# количество моделей в ответе
# ошибки больше нуля, р2 меньше единицы
# итого 9

# scaler
# нормализация
# параметры

# basktest
# покупка продажа



import numpy as np

class Scaler:
    '''
    Represents default minmax scaler adapted for price scaling.
    
    Methods:
    
    fit(array) - defines minimum and maximum values for scaling;
    
    scale(array) - scales array by defined minimum and maximum values;
    
    fit_scale(fit_array, scale_array) - defines minimum and maximum from fit_array, then scales fit_array and scale_array using defined parameters;
    
    unscale(array) - scales array back to source range;
    
    params() - returns minimum and maximum parameters of scaling.
    '''
    
    minimum, maximum = 0, 0
    
    def __init__(self, params = None):
        if params != None:
            self.minimum, self.maximum = params[0], params[1]

    def fit(self, array_to_fit) -> any:
        '''
        Fits scaler on input array.
        '''
        self.minimum = np.min(array_to_fit)
        self.maximum = np.max(array_to_fit)
        return self
    
    def scale(self, array) -> np.array:
        '''
        Returns array scaled on data fed during fit.
        '''
        return (array-self.minimum)/(self.maximum - self.minimum)
    
    def fit_scale(self, fit_array, scale_array) -> list:
        '''
        Returns two arrays, scaled on first array range.
        '''
        self.minimum = np.min(fit_array)
        self.maximum = np.max(fit_array)
        return (fit_array-self.minimum)/(self.maximum - self.minimum), (scale_array-self.minimum)/(self.maximum - self.minimum)
    
    def unscale(self, array) -> np.array:
        '''
        Returns array, unscaled to source range.
        '''
        return array*(self.maximum - self.minimum) + self.minimum
    
    def params(self) -> list:
        '''
        Returns min and max values, used to scale.
        '''
        return self.minimum, self.maximum


# Data processing
import numpy as np, pandas as pd, time

# ML methods
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import RandomForestRegressor, GradientBoostingRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.svm import SVR
import catboost as cb
import xgboost as xg

# Autoregressive methods
from statsmodels.tsa.arima.model import ARIMA

# NN methods
from tensorflow.keras import Sequential, models
from tensorflow.keras.layers import LSTM, Dense, GRU, SimpleRNN, Conv1D, Flatten

# Metrics
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score,  mean_absolute_percentage_error

# Muting warnings
import warnings
warnings.filterwarnings("ignore")

from .preprocessing import get_data
from .scaler import Scaler


from pathlib import Path
PACKAGEDIR = Path(__file__).parent.absolute()

"""
The mod module
"""


# def get_topologies(lag = 30):
#     neurons_per_layer = lag-1
#     input_shape = (neurons_per_layer, 1)
#     return {  'CNN + LSTM': [Conv1D(filters = 128, kernel_size = 3, activation = 'relu', input_shape = input_shape),
#                                 LSTM(units = neurons_per_layer),
#                                 Dense(units = 1)],
#                     'LSTM x3': [LSTM(units = neurons_per_layer, 
#                                     return_sequences = True, 
#                                     input_shape = input_shape),
#                                 LSTM(units = neurons_per_layer, 
#                                     return_sequences = True),
#                                 LSTM(units = neurons_per_layer),
#                                 Dense(units = 1)],
#                     'LSTM x2': [LSTM(units = neurons_per_layer, 
#                                     return_sequences = True, 
#                                     input_shape = input_shape),
#                                 LSTM(units = neurons_per_layer),
#                                 Dense(units = 1)],
#                     'LSTM x1': [LSTM(units = neurons_per_layer, 
#                                     input_shape = input_shape),
#                                 Dense(units = 1)],
#                     'CNN + GRU': [Conv1D(filters = 128, kernel_size = 3, activation = 'relu', input_shape = input_shape),
#                                 GRU(units = neurons_per_layer),
#                                 Dense(units = 1)],
#                     'GRU x3' : [GRU(units = neurons_per_layer, 
#                                     return_sequences = True, 
#                                     input_shape = input_shape),
#                                 GRU(units = neurons_per_layer, 
#                                     return_sequences = True),
#                                 GRU(units = neurons_per_layer),
#                                 Dense(units = 1)],
#                     'GRU x2' : [GRU(units = neurons_per_layer, 
#                                     return_sequences = True, 
#                                     input_shape = input_shape),
#                                 GRU(units = neurons_per_layer),
#                                 Dense(units = 1)],
#                     'GRU x1' : [GRU(units = neurons_per_layer, 
#                                     input_shape = input_shape),
#                                 Dense(units = 1)],
#                     'CNN + SimpleRNN': [Conv1D(filters = 128, kernel_size = 3, activation = 'relu', input_shape = input_shape),
#                                         SimpleRNN(units = neurons_per_layer),
#                                         Dense(units = 1)],
#                     'SimpleRNN x3':[SimpleRNN(units = neurons_per_layer, 
#                                             return_sequences = True, 
#                                             input_shape = input_shape),
#                                     SimpleRNN(units = neurons_per_layer, 
#                                             return_sequences = True),
#                                     SimpleRNN(units = neurons_per_layer),
#                                     Dense(units = 1)],
#                     'SimpleRNN x2':[SimpleRNN(units = neurons_per_layer, 
#                                             return_sequences = True, 
#                                             input_shape = input_shape),
#                                     SimpleRNN(units = neurons_per_layer),
#                                     Dense(units = 1)],
#                     'SimpleRNN x1':[SimpleRNN(units = neurons_per_layer, 
#                                             input_shape = input_shape),
#                                     Dense(units = 1)],
#                     'CNN': [Conv1D(filters = 32, kernel_size = 5, input_shape = input_shape, activation = 'relu'),
#                             Flatten(),
#                             Dense(units = 1)],
#                     'MLP(3)': [Dense(units = neurons_per_layer, input_shape = (neurons_per_layer,)),
#                             Dense(units = neurons_per_layer*2),
#                             Dense(units = neurons_per_layer),
#                             Dense(units = 1)],        
#                     'MLP(2)': [Dense(units = neurons_per_layer, input_shape = (neurons_per_layer,)),
#                             Dense(units = neurons_per_layer),
#                             Dense(units = 1)],
#                     'MLP(1)': [Dense(units = neurons_per_layer, input_shape = (neurons_per_layer,)),
#                             Dense(units = 1)],
#                     }

# def evaluate_ml_models(x_train, x_test, y_train, y_test, scaler):
#     '''
#     Evaluates LR, DTR, RFR, GBR and SVR models on data.
#     '''
#     statistics = []
#     models = {  'LR' :  LinearRegression(),
#                 'DTR' : DecisionTreeRegressor(min_samples_leaf=5),
#                 'RFR' : RandomForestRegressor(),
#                 'GBR' : GradientBoostingRegressor(),
#                 'SVR' : SVR(kernel='linear', epsilon=1e-3)}
    
#     for model_description, model in models.items():
#         start_time = time.time()
#         model.fit(x_train, y_train)
#         utilized = time.time() - start_time
#         preds = model.predict(x_test)
#         statistics.append({ 'time': utilized, 
#                             'mse': mean_squared_error(scaler.unscale(y_test), scaler.unscale(preds)),
#                             'mae': mean_absolute_error(scaler.unscale(y_test), scaler.unscale(preds)),
#                             'mape': mean_absolute_percentage_error(scaler.unscale(y_test), scaler.unscale(preds)),
#                             'r2':  r2_score(scaler.unscale(y_test), scaler.unscale(preds)),
#                             'model': model_description})
#     cbr = cb.CatBoostRegressor(loss_function='MAPE')
#     cbr.fit(x_train,y_train,silent=True)
#     xgbr = xg.XGBRegressor(objective='reg:squarederror')
#     xgbr.fit(x_train,y_train)
#     xgbrfr = xg.XGBRFRegressor(objective = 'reg:squarederror')
#     xgbrfr.fit(x_train,y_train)
#     for j,i in enumerate([cbr,xgbr,xgbrfr]):
#         preds = i.predict(x_test)
#         statistics.append({ 'time': utilized, 
#                     'mse': mean_squared_error(scaler.unscale(y_test), scaler.unscale(preds)),
#                     'mae': mean_absolute_error(scaler.unscale(y_test), scaler.unscale(preds)),
#                     'mape': mean_absolute_percentage_error(scaler.unscale(y_test), scaler.unscale(preds)),
#                     'r2':  r2_score(scaler.unscale(y_test), scaler.unscale(preds)),
#                     'model': ['CBR', 'XGBR', 'XGBRFR'][j]})
#     return statistics   


# def evaluate_neural_networks(topologies_dict: dict, x_train, x_test, y_train, y_test, scaler, optimizer = 'nadam', loss = 'mse', epochs = 10, batch_size = 256) -> list:
#     '''
#     Evaluates neural networks with defined topologies with defined hyperparameters on provided data.
#     '''
#     statistics = []
#     for topology_description, topology in topologies_dict.items():
#         current_topology_model = Sequential(topology)
#         current_topology_model.compile(optimizer = optimizer, loss = loss, metrics = ['mae', 'mse'])
#         start_time = time.time()
#         current_topology_model.fit(x = x_train, y = y_train, epochs = epochs, batch_size = batch_size, verbose = 0)
#         utilized_time = time.time() - start_time
#         preds = current_topology_model.predict(x_test, verbose=0).reshape(-1)
#         statistics.append({'time':utilized_time, 
#                     'mse': mean_squared_error(scaler.unscale(y_test), scaler.unscale(preds)),
#                     'mae': mean_absolute_error(scaler.unscale(y_test), scaler.unscale(preds)),
#                     'mape': mean_absolute_percentage_error(scaler.unscale(y_test), scaler.unscale(preds)),
#                     'r2':  r2_score(scaler.unscale(y_test), scaler.unscale(preds)),
#                     'model': topology_description})
#     return statistics

# def evaluate_autoregressive_models(x_train, x_test, y_train, y_test, scaler) -> list:
#     '''
#     Evaluates ARMA and ARIMA models on provided data.
#     '''
#     statistics = []
#     p, q = 2, 1
#     for d in range(2):
#         forecasts, true = [], []
#         for i in range(int(len(y_test)/10)):
#             start_time = time.time()
#             model = ARIMA(scaler.unscale(x_test[i]), order = (p, d, q), enforce_stationarity=False)
#             forecasts.append(model.fit().forecast(steps = 1)[0])
#             true.append(scaler.unscale(y_test[i]))
#             utilized = time.time() - start_time
#         statistics.append({'time':utilized, 
#                     'mse': mean_squared_error(np.array(true), np.array(forecasts)),
#                     'mae': mean_absolute_error(np.array(true), np.array(forecasts)),
#                     'mape': mean_absolute_percentage_error(np.array(true), np.array(forecasts)),
#                     'r2':  r2_score(np.array(true), np.array(forecasts)),
#                     'model': f'ARIMA({p},{d},{q})' if d > 0 else f'ARMA({p},{q})'})
#     return statistics

# def evaluate_all_models(tickers = [], length = 15):
#     '''
#     Evaluates ml, ar and nn
#     '''
#     for i in tickers:

#         flat_x_train, flat_x_test, flat_y_train, flat_y_test, flat_scaler = get_data(ticker = i, length = length, flatten = True)
#         statistics = evaluate_ml_models(flat_x_train, flat_x_test, flat_y_train, flat_y_test, flat_scaler)
#         statistics +=  evaluate_autoregressive_models(flat_x_train, flat_x_test, flat_y_train, flat_y_test, flat_scaler)

#         x_train, x_test, y_train, y_test, scaler = get_data(ticker = i, length = length)
#         statistics += evaluate_neural_networks(get_topologies(length), x_train, x_test, y_train, y_test, scaler)

#         current_ticker_statistics = pd.DataFrame(statistics)
#         current_ticker_statistics['ticker'] = i
#         total_statistics = pd.concat([total_statistics, current_ticker_statistics])
#     return pd.DataFrame(statistics).sort_values(by='mape')


import pandas as pd, time
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score,  mean_absolute_percentage_error
import plotly.express as px
from .model import Model

import warnings
warnings.filterwarnings("ignore")


def evaluate_ml_models(x_train, x_test, y_train, y_test, scaler, out_type = 'list'):
    '''
    Evaluates machine learning models on data.

    Args:
        x_train (np.array): Input values to fit model.
        x_test (np.array): Input values to evaluate model.
        y_train (np.array): Target values to fit model.
        y_test (np.array): Target values to fit model.
        scaler (Scaler): Scaler to scale/unscale values during evaluation.
        out_type (str): Format to return data. 
    
    Returns:
        list: List of dictionaries with time, MSE, MAE, MAPE, R2 values and model name for each model.
        or
        pd.DataFrame : dataframe with the same data.

    Example:
    >>> evaluate_ml_models(x_train, x_test, y_train, y_test, scaler)
    [{'time' : 0.12, 'mse' : 0.04, 'mae' : 0.2, 'mape' : 0.01, 'model' : 'LR'}, ...]
    '''
    statistics = []
    for i in ['LR', 'DTR', 'RFR', 'GBR', 'SVR', 'CBR', 'XGBR', 'XGBRFR']:
        current_model = Model(i)
        start_time = time.time()
        current_model.fit(x_train, y_train)
        utilized = time.time() - start_time
        pred = current_model.predict(x_test, scaler)
        true = scaler.unscale(y_test)
        statistics.append({ 'time': utilized, 
                    'mse': mean_squared_error(true, pred),
                    'mae': mean_absolute_error(true, pred),
                    'mape': mean_absolute_percentage_error(true, pred),
                    'r2':  r2_score(true, pred),
                    'model': i})
        
    if out_type == 'df':
        return pd.DataFrame(statistics).sort_values(by='mape')
    elif out_type == 'list':
        return statistics

def evaluate_ar_models(x_train, x_test, y_train, y_test, scaler, out_type = 'list'):
    '''
    Evaluates autoregressive models on data.

    Args:
        x_train (np.array): Input values to fit model.
        x_test (np.array): Input values to evaluate model.
        y_train (np.array): Target values to fit model.
        y_test (np.array): Target values to fit model.
        scaler (Scaler): Scaler to scale/unscale values during evaluation.
        out_type (str): Format to return data. 
    
    Returns:
        list: List of dictionaries with time, MSE, MAE, MAPE, R2 values and model name for each model.
        or
        pd.DataFrame : dataframe with the same data.

    Example:
    >>> evaluate_ar_models(x_train, x_test, y_train, y_test, scaler)
    [{'time' : 0.12, 'mse' : 0.04, 'mae' : 0.2, 'mape' : 0.01, 'model' : 'ARMA(2,1)'}, ...]
    '''
    statistics = []
    for i in ['ARMA(2,1)', 'ARIMA(2,1,1)']:
        pred = []
        for j in range(len(y_test)):
            start_time = time.time()
            current_model = Model(i, data=x_test[j])            
            utilized = time.time() - start_time
            pred.append(current_model.predict(x_test[j], scaler))
        true = scaler.unscale(y_test)
        statistics.append({ 'time': utilized, 
                    'mse': mean_squared_error(true, pred),
                    'mae': mean_absolute_error(true, pred),
                    'mape': mean_absolute_percentage_error(true, pred),
                    'r2':  r2_score(true, pred),
                    'model': i})
        
    if out_type == 'df':
        return pd.DataFrame(statistics).sort_values(by='mape')
    elif out_type == 'list':
        return statistics  

def evaluate_nn_models(x_train, x_test, y_train, y_test, scaler, optimizer = 'nadam', loss = 'mse', epochs = 10, batch_size = 256, out_type = 'list'):
    '''
    Evaluates neural network based models on data.

    Args:
        x_train (np.array): Input values to fit model.
        x_test (np.array): Input values to evaluate model.
        y_train (np.array): Target values to fit model.
        y_test (np.array): Target values to fit model.
        scaler (Scaler): Scaler to scale/unscale values during evaluation.
        optimizer (str): tf.Keras optimizer name (optional: only for neural networks).
        loss (str): tf.Keras loss name (optional: only for neural networks).
        epochs (int): Number of epochs during fitting (optional: only for neural networks).
        batch_size (int): Size of batch during fitting (optional: only for neural networks).
        out_type (str): Format to return data. 
    
    Returns:
        list: List of dictionaries with time, MSE, MAE, MAPE, R2 values and model name for each model.
        or
        pd.DataFrame : dataframe with the same data.

    Example:
    >>> evaluate_nn_models(x_train, x_test, y_train, y_test, scaler)
    [{'time' : 0.12, 'mse' : 0.04, 'mae' : 0.2, 'mape' : 0.01, 'model' : 'LSTM x1'}, ...]
    '''

    statistics = []
    for i in ['CNN + LSTM', 'LSTM x3', 'LSTM x2', 'LSTM x1', 'CNN + GRU', 'GRU x3', 'GRU x2', 'GRU x1', \
              'CNN + SimpleRNN', 'SimpleRNN x3', 'SimpleRNN x2', 'SimpleRNN x1', 'CNN', 'MLP(3)', 'MLP(2)','MLP(1)']:
        current_model = Model(i, lag = x_train.shape[1]+1, optimizer=optimizer, loss=loss)
        start_time = time.time()
        current_model.fit(x_train, y_train, epochs, batch_size)
        utilized = time.time() - start_time
        pred = current_model.predict(x_test, scaler)
        true = scaler.unscale(y_test)
        statistics.append({ 'time': utilized, 
                    'mse': mean_squared_error(true, pred),
                    'mae': mean_absolute_error(true, pred),
                    'mape': mean_absolute_percentage_error(true, pred),
                    'r2':  r2_score(true, pred),
                    'model': i})
    
    if out_type == 'df':
        return pd.DataFrame(statistics).sort_values(by='mape')
    elif out_type == 'list':
        return statistics


def evaluate_all_models(x_train, x_test, y_train, y_test, scaler, optimizer = 'nadam', loss = 'mse', epochs = 10, batch_size = 256, out_type = 'df'):
    '''
    Evaluates all models on data.

    Args:
        x_train (np.array): Input values to fit model.
        x_test (np.array): Input values to evaluate model.
        y_train (np.array): Target values to fit model.
        y_test (np.array): Target values to fit model.
        scaler (Scaler): Scaler to scale/unscale values during evaluation.
        optimizer (str): tf.Keras optimizer name (optional: only for neural networks).
        loss (str): tf.Keras loss name (optional: only for neural networks).
        epochs (int): Number of epochs during fitting (optional: only for neural networks).
        batch_size (int): Size of batch during fitting (optional: only for neural networks).
        out_type (str): Format to return data. 
    
    Returns:
        list: List of dictionaries with time, MSE, MAE, MAPE, R2 values and model name for each model.
        or
        pd.DataFrame : dataframe with the same data.

    Example:
    >>> evaluate_all_models(x_train, x_test, y_train, y_test, scaler)
    [{'time' : 0.12, 'mse' : 0.04, 'mae' : 0.2, 'mape' : 0.01, 'model' : 'LR'}, ...]
    '''

    statistics = evaluate_ml_models(x_train, x_test, y_train, y_test, scaler)
    statistics += evaluate_ar_models(x_train, x_test, y_train, y_test, scaler)
    statistics += evaluate_nn_models(x_train, x_test, y_train, y_test, scaler, optimizer, loss, epochs, batch_size)

    if out_type == 'df':
        return pd.DataFrame(statistics).sort_values(by='mape')
    elif out_type == 'list':
        return statistics
    
def plot_histogran(results):
    '''
    Plots histogram for resulting error values.
    '''
    px.histogram(results, x='model', y='mape')